ctypedef PatternMap[unsigned int,BaseValueHandler[uint],unsigned long long] t_relationmap
ctypedef PatternMap[double,BaseValueHandler[double],unsigned long long] t_relationmap_double
ctypedef PatternMap[unsigned int,BaseValueHandler[uint],unsigned long long].iterator t_relationmap_iterator
ctypedef PatternMap[double,BaseValueHandler[double],unsigned long long].iterator t_relationmap_double_iterator
