191	pangs
189	Th'oppressor's
188	whose
187	scorns
185	long
184	so
183	puzzles
182	calamity
181	contumely
180	respect
179	heart-ache
178	bare
177	undiscovered
176	coil
175	pitch
174	remember'd
171	may
170	dreams
169	there's
168	whips
167	outrageous
166	Ay
165	dream
164	There's
163	wish'd
162	nobler
161	himself
159	consummation
158	mortal
157	Or
156	perchance
155	off
153	flesh
152	slings
150	opposing
149	thought
147	shocks
146	arrows
144	fortune
143	sea
142	orisons
141	give
140	sins
139	shuffled
138	takes
137	them
135	thousand
134	traveller
128	Must
127	mind
126	cowards
125	fair
124	suffer
123	rub
121	natural
120	heir
119	those
118	more
117	question
116	pause
114	Whether
113	delay
112	say
111	arms
105	troubles
103	despised
102	quietus
186	against
48	would
47	who
19	see
46	makes
53	Soft
160	Devoutly
41	life
87	his
42	death
40	'tis
38	No
190	wrong
49	by
122	hue
26	bear
70	fly
36	die
11	To
50	's
34	For
84	grunt
30	?
62	pale
172	come
43	all
51	thy
145	man
39	this
148	bodkin
15	The
8	to
115	know
35	When
29	pee
132	what
131	Thus
130	awry
12	not
77	bourn
80	after
10	.
110	proud
109	returns
108	their
107	Ophelia
32	!
129	patient
14	or
133	insolence
9	the
90	he
45	With
28	us
13	and
7	of
6	,
173	take
44	have
31	make
16	And
33	end
136	time,
24	flee
106	native
21	we
151	thus
27	That
17	a
37	is
23	be
154	—
18	sleep
104	conscience
20	;
25	in
63	sicklied
72	Than
71	great
22	that
52	you
54	name
81	But
55	turn
75	rather
58	regard
56	Nymph
57	action
59	moment
60	enterprises
85	my
86	fardels
61	cast
64	lose
65	Is
66	now
67	does
68	o'er
69	others
74	ills
73	resolution
76	will
78	from
91	unworthy
92	th'
79	country
82	under
83	sweat
96	spurns
89	might
88	currents
93	something
94	weary
95	merit
98	dread
97	with
99	office
100	law
101	love
